package{
  import flash.display.Graphics;
  import flash.display.Sprite;
  import flash.geom.Matrix3D;
  import flash.geom.Vector3D;
  import flash.text.TextField;
  import flash.text.TextFormat;
  import component.TextBox;
  import component.ViewModel;
  import figure.Cube3D;
  import figure.Point3D;

//--------------------------------------------------------------------//
// ̌nf[^\f
  public class MBDViewModel extends ViewModel{

// ̍ŏl
    public static const MIN_VALUE:Number=1.0E-20;

// ̌nf[^f
    protected var model:MultiBodyModel;

// ]
    internal var axis:Vector3D;

// Wϊ}gbNX
    internal var trans:Matrix3D;

// W
    internal var coordinates:Coordinates;

// f\̈TCY
    protected var viewSize:Number;

// fsړW
    internal var transCoef:Number;

// f]xW
    internal var rotateCoef:Number;

// fgEkW
    internal var zoomCoef:Number;

// OiF
    internal var fgColor:uint;

// wiF
    internal var bgColor:uint;

// \
    internal var time:TextField;

// ̕`惊Xg
    internal var drawBody:Vector.<DrawBody>;

// _`W
    internal var vp0:Vector3D;

//--------------------------------------------------------------------//
// ̌nf[^\f\z
// sprite - eXvCg
// model ̌nf[^f
// graphics - `OtBbNX
// x,y,width,height - \ʒuAA
// fgColor - OiF
// bgColor - wiF
// textFormat - eLXg
    public function MBDViewModel(sprite:Sprite,model:MultiBodyModel,
      	      	      	      	 graphics:Graphics,x:int,y:int,
      	      	      	      	 width:int,height:int,
      	      	      	      	 fgColor:uint,bgColor:uint,
      	      	      	      	 textFormat:TextFormat){
      super(graphics,x,y,width,height);
      this.model=model;
      trans=new Matrix3D();
      axis=new Vector3D();
      viewSize=Math.min(width,height);
      coordinates=new Coordinates(0.2*viewSize,textFormat);
      transCoef=1.0;
      rotateCoef=180.0/viewSize;
      zoomCoef=1.0/viewSize;
      this.fgColor=fgColor;
      this.bgColor=bgColor;

      time=new TextField();
      time.defaultTextFormat=textFormat;
      time.background=true;
      time.backgroundColor=bgColor;
      time.selectable=false;
      time.x=10;
      time.y=10;

      sprite.addChild(coordinates);
      sprite.addChild(time);
      setDraw();
    }

//--------------------------------------------------------------------//
// ̌nf[^fԂ
    public function getModel():MultiBodyModel{
      return model;
    }

//--------------------------------------------------------------------//
// f𕽍sړ
// dx,dy - ʏ̈ړ
    public function translate(dx:int,dy:int):void{
      trans.appendTranslation(transCoef*dx,-transCoef*dy,0.0);
    }

//--------------------------------------------------------------------//
// f]
// dx,dy - ʏ̈ړ
    public function rotate(dx:int,dy:int):void{
      var r:Number=Math.sqrt(dx*dx+dy*dy);
      axis.x=dy/r;
      axis.y=dx/r;
      r*=rotateCoef;
      trans.appendRotation(r,axis);
      coordinates.trans.appendRotation(r,axis);
    }

//--------------------------------------------------------------------//
// gEk
// dx,dy - ʏ̈ړ
    public function zoom(dx:int,dy:int):void{
      var c:Number=Math.pow(2.0,zoomCoef*Number(dx+dy));
      trans.appendScale(c,c,c);
      coordinates.trans.appendScale(c,c,c);
    }

//--------------------------------------------------------------------//
// `ʒu
// z
    public function initView():void{
      trans.identity();
      coordinates.trans.identity();
      draw();
    }

//--------------------------------------------------------------------//
// xɂ
    public function viewX():void{
      trans.identity();
      trans.appendRotation(-90.0,Vector3D.Y_AXIS);
      trans.appendRotation(-90.0,Vector3D.Z_AXIS);
      coordinates.trans.identity();
      coordinates.trans.appendRotation(-90.0,Vector3D.Y_AXIS);
      coordinates.trans.appendRotation(-90.0,Vector3D.Z_AXIS);
      draw();
    }

//--------------------------------------------------------------------//
// yɂ
    public function viewY():void{
      trans.identity();
      trans.appendRotation(-90.0,Vector3D.X_AXIS);
      coordinates.trans.identity();
      coordinates.trans.appendRotation(-90.0,Vector3D.X_AXIS);
      draw();
    }

//--------------------------------------------------------------------//
// `ݒ
    public function setDraw():void{
      var bounds:Cube3D=model.getBounds();
      var center:Point3D=bounds.getCenterLocation();
      var boundSize:Number=Math.max(Math.max(bounds.width,
      	      	      	      	    bounds.height),bounds.depth);
      var vCoef:Number=0.8*viewSize/Math.max(boundSize,MIN_VALUE);
      vp0=new Vector3D(-vCoef*center.x,-vCoef*center.y,
      	      	       -vCoef*center.z,1.0);
      var bodyCount:int=model.getBodyCount();
      drawBody=new Vector.<DrawBody>(bodyCount,true);
      for(var i:int=0;i<bodyCount;i++){
      	drawBody[i]=new DrawBody(model.getBody(i),vCoef,center);
      }
    }

//--------------------------------------------------------------------//
// `悷
    public function draw():void{
      drawInit(bgColor);
      var vp:Vector3D=trans.transformVector(vp0);
      graphics.lineStyle(1,fgColor);
      coordinates.draw(graphics,cx+vp.x,cy-vp.y);

      for(var i:int=0;i<drawBody.length;i++){
      	drawBody[i].draw(graphics,trans,cx,cy,fgColor);
      }
      time.text=timeString(model.getTime().time());
      time.width=time.textWidth+TextBox.CL;
      time.height=time.textHeight+TextBox.CL;
    }

//--------------------------------------------------------------------//
// ̕\Ԃ
// t - 
    internal function timeString(t:Number):String{
      if(t>=1.0E5){
      	return "Time = "+String(t);
      }
      else if((t>=1.0) || (t==0.0)){
      	return "Time = "+TextBox.format(t,3);
      }
      else if(t>=0.001){
      	return "Time = "+TextBox.format(t,4);
      }
      else{
      	return "Time = "+String(t);
      }
    }
  }
}
