package{
  import flash.geom.Vector3D;
  import constraint.Spring;
  import figure.Cube3D;
  import figure.DefaultMatrix;
  import figure.DefaultMatrixSolver;
  import figure.Point3D;

//--------------------------------------------------------------------//
// ̌nf[^f
  public class MultiBodyModel{

// Runge - Kutta @̌W1
    public static const C1:Vector.<Number>=
      Vector.<Number>([1.0/6.0,1.0/3.0,1.0/3.0,1.0/6.0]);

// Runge - Kutta @̌W2
    public static const C2:Vector.<Number>=
      Vector.<Number>([0.5,0.5,1.0]);

// ̃Xg
    protected var bodies:Vector.<Body>;

// SXg
    protected var rigids:Vector.<RigidConstraint>;

// _SXg
    protected var springs:Vector.<Constraint>;

// }gbNXExNg̃TCY
    protected var size:int;

// ʁES}gbNX
    protected var matrix:DefaultMatrixSolver;

// ׏dxNg
    protected var vector:Vector.<Number>;

// Runge-Kutta @Xebṽf[^
    protected var tempPos:Array;

// fE
    public var bounds:Cube3D;

// 
    protected var time:Time;

// d͉xxNg
    protected var g:Vector3D;

//--------------------------------------------------------------------//
// ̌nf[^f\z
    public function MultiBodyModel(){
      bodies=new Vector.<Body>();
      rigids=new Vector.<RigidConstraint>();
      springs=new Vector.<Constraint>();
      size=0;
      bounds=new Cube3D();
      time=new Time(1.0);
      g=new Vector3D(0.0,0.0,-9.80665,1.0);
    }

//--------------------------------------------------------------------//
// fE\̂Ԃ
    public function getBounds():Cube3D{
      return bounds;
    }

//--------------------------------------------------------------------//
// ԂԂ
    public function getTime():Time{
      return time;
    }

//--------------------------------------------------------------------//
// ԍݕݒ肷
// dt - ԍݕ
    public function setDt(dt:Number):void{
      time=new Time(dt);
    }

//--------------------------------------------------------------------//
// d͉xݒ肷
// g - d͉xxNg
    public function setGravity(g:Vector3D):void{
      this.g=g;
    }

//--------------------------------------------------------------------//
// ̂̐Ԃ
    public function getBodyCount():int{
      return bodies.length;
    }

//--------------------------------------------------------------------//
// ̂Ԃ
// index - ̂̃CfbNX
    public function getBody(index:int):Body{
      return bodies[index];
    }

//--------------------------------------------------------------------//
// ̂ǉ
// body - 
    public function addBody(body:Body):void{
      bodies.push(body);
    }

//--------------------------------------------------------------------//
// S̐Ԃ
    public function getConstraintCount():int{
      return rigids.length;
    }

//--------------------------------------------------------------------//
// SԂ
// index - S̃CfbNX
    public function getConstraint(index:int):RigidConstraint{
      return rigids[index];
    }

//--------------------------------------------------------------------//
// Sǉ
// constraint - S
    public function addConstraint(constraint:RigidConstraint):void{
      rigids.push(constraint);
    }

//--------------------------------------------------------------------//
// _S̐Ԃ
    public function getSpringCount():int{
      return springs.length;
    }

//--------------------------------------------------------------------//
// _SԂ
// index - S̃CfbNX
    public function getSpring(index:int):Constraint{
      return springs[index];
    }

//--------------------------------------------------------------------//
// _Sǉ
// spring - S
    public function addSpring(spring:Spring):void{
      springs.push(spring);
    }

//--------------------------------------------------------------------//
// f[^
    public function clear():void{
      bodies.length=0;
      rigids.length=0;
      springs.length=0;
      size=0;
      bounds=new Cube3D();
    }

//--------------------------------------------------------------------//
// f
    public function init():void{
      bodies.sort(sortLabel);
      reNumbering();
      rigids.sort(sortConstraint);
      springs.sort(sortConstraint);
      setSize();
      setTempPosition1();
      setGcPosition();
      setInitPosition();
      setBounds();
    }

//--------------------------------------------------------------------//
// ߓ_|C^ݒ肷
    public function reNumbering():void{
      var labelMap:LabelMap=new LabelMap();
      var bodyCount:int=getBodyCount();
      for(var i:int=0;i<bodyCount;i++){
      	labelMap.put(getBody(i),i);
      }
      var constraintCount:int=getConstraintCount();
      for(i=0;i<constraintCount;i++){
      	labelMap.resetConstraint(getConstraint(i));
      }
      var springCount:int=getSpringCount();
      for(i=0;i<springCount;i++){
      	labelMap.resetConstraint(getSpring(i));
      }
      labelMap.clear();
    }

//--------------------------------------------------------------------//
// }gbNXExNg̃TCYݒ肷
    protected function setSize():void{
      size=6*getBodyCount();
      var constraintCount:int=getConstraintCount();
      for(var i:int=0;i<constraintCount;i++){
      	size+=getConstraint(i).dof();
      }
    }

//--------------------------------------------------------------------//
// Runge-Kutta @Xebṽf[^ݒ肷
    protected function setTempPosition1():void{
      var bodyCount:int=getBodyCount();
      tempPos=new Array(bodyCount);
      for(var i:int=0;i<bodyCount;i++){
      	tempPos[i]=new Vector.<Position>(4,true);
      	tempPos[i][0]=getBody(i).position();
      }
    }

//--------------------------------------------------------------------//
// dSʒuExݒ肷
    protected function setGcPosition():void{
      var constraintCount:int=getConstraintCount();
      for(var i:int=0;i<constraintCount;i++){
      	var c:RigidConstraint=getConstraint(i);
      	c.setGcPosition(getPosition(c.p1(),0),getPosition(c.p2(),0));
      }
    }

//--------------------------------------------------------------------//
// ʒuݒ肷
    protected function setInitPosition():void{
      time.init();
      var bodyCount:int=getBodyCount();
      for(var i:int=0;i<bodyCount;i++){
      	getBody(i).setInitPosition();
      }
    }

//--------------------------------------------------------------------//
// fEݒ肷
    protected function setBounds():void{
      bounds.setBounds();
      var bodyCount:int=getBodyCount();
      if(bodyCount==0) return;
      for(var i:int=0;i<bodyCount;i++){
      	var pt:Array=getBody(i).vertices();
      	for(var i1:int=0;i1<pt.length;i1++){
      	  for(var j1:int=0;j1<pt[i1].length;j1++){
      	    if((i==0) && (i1==0) && (j1==0)){
      	      bounds.setLocation(pt[i1][j1].x,pt[i1][j1].y,
      	      	      	      	 pt[i1][j1].z);
      	    }
      	    else{
      	      bounds.inflate(pt[i1][j1].x,pt[i1][j1].y,pt[i1][j1].z);
      	    }
      	  }
      	}
      }
    }

//--------------------------------------------------------------------//
// ʒuԂɖ߂
    public function initPosition():void{
      time.init();
      var bodyCount:int=getBodyCount();
      for(var i:int=0;i<bodyCount;i++){
      	getBody(i).initPosition();
      }
      setTempPosition1();
      setGcPosition();
    }

//--------------------------------------------------------------------//
// oEKë艻p[^ݒ肷
// alpha - oEKë艻p[^
// beta - oEKë艻p[^
    public function setBaumgarte(alpha:Number,beta:Number):void{
      var constraintCount:int=getConstraintCount();
      for(var i:int=0;i<constraintCount;i++){
      	getConstraint(i).setBaumgarte(alpha,beta);
      }
    }

//--------------------------------------------------------------------//
// ̃XebvvZ
    public function next():void{
      time.next();
      var dt:Number=time.dt();
      var tol:Number=1.0e-10;
      for(var i:int=0;i<4;i++){
      	setTempPosition2(dt,i);
      	createMatrixVector(i);
      	setAcc(matrix.solveByLUNoPreserv(vector,tol),i);
      }
      update(dt);
    }

//--------------------------------------------------------------------//
// Runge-Kutta @Xebṽf[^ݒ肷
// dt - ԍݕ
// step - Runge-Kutta@̃Xebv
    protected function setTempPosition2(dt:Number,step:int):void{
      if(step==0) return;
      var cdt:Number=dt*C2[step-1];
      var bodyCount:int=getBodyCount();
      for(var i:int=0;i<bodyCount;i++){
      	tempPos[i][step]=getBody(i).position().copy();
      	tempPos[i][step].update2(cdt,tempPos[i][step-1].v,
      	      	      	      	 tempPos[i][step-1].a,
      	      	      	      	 tempPos[i][step-1].eulerVel(),
      	      	      	      	 tempPos[i][step-1].dw);
      }
    }

//--------------------------------------------------------------------//
// ʁES}gbNXщ׏dxNg쐬
// step - Runge-Kutta@̃Xebv
    protected function createMatrixVector(step:int):void{
      matrix=new DefaultMatrixSolver(size);
      vector=new Vector.<Number>(size,true);

      var ii:int=0;
      var bodyCount:int=getBodyCount();
      for(var i:int=0;i<bodyCount;i++){
      	var body:Body=getBody(i);
      	var mg:Vector3D=new Vector3D(body.mass*g.x,body.mass*g.y,
      	      	      	      	     body.mass*g.z);
      	var m:DefaultMatrix=body.massMatrix();
      	var f:Vector.<Number>=body.forceVector
      	  (mg,new Vector3D(),getPosition(i,step));
      	for(var i1:int=0;i1<6;i1++){
      	  for(var j1:int=0;j1<6;j1++){
      	    matrix.set(ii+i1,ii+j1,m.get(i1,j1));
      	  }
      	  vector[ii+i1]=f[i1];
      	}
      	ii+=6;
      }

      var constraintCount:int=getConstraintCount();
      for(i=0;i<constraintCount;i++){
      	var c:RigidConstraint=getConstraint(i);
      	var dof:int=c.dof();
      	c.create(getPosition(c.p1(),step),getPosition(c.p2(),step));
      	var mm:Vector.<DefaultMatrix>=c.getMatrix();
      	f=c.getVector();
      	if(c.p1()>=0){
      	  var j0:int=6*c.p1();
      	  for(i1=0;i1<dof;i1++){
      	    for(j1=0;j1<6;j1++){
      	      matrix.set(ii+i1,j0+j1,mm[0].get(i1,j1));
      	      matrix.set(j0+j1,ii+i1,mm[0].get(i1,j1));
      	    }
      	  }
      	}
      	if(c.p2()>=0){
      	  j0=6*c.p2();
      	  for(i1=0;i1<dof;i1++){
      	    for(j1=0;j1<6;j1++){
      	      matrix.set(ii+i1,j0+j1,mm[1].get(i1,j1));
      	      matrix.set(j0+j1,ii+i1,mm[1].get(i1,j1));
      	    }
      	  }
      	}
      	for(i1=0;i1<dof;i1++){
      	  vector[ii+i1]=f[i1];
      	}
      	ii+=dof;
      }

      var springCount:int=getSpringCount();
      for(i=0;i<springCount;i++){
      	var s:Constraint=getSpring(i);
      	dof=s.dof();
      	s.create(getPosition(s.p1(),step),getPosition(s.p2(),step));
      	f=s.getVector();
      	if(s.p1()>=0){
      	  var i0:int=6*s.p1();
      	  for(i1=0;i1<dof;i1++){
      	    vector[i0+i1]+=f[i1];
      	  }
      	}
      	if(s.p2()>=0){
      	  i0=6*s.p2();
      	  for(i1=0;i1<dof;i1++){
      	    vector[i0+i1]+=f[dof+i1];
      	  }
      	}
      }
    }

//--------------------------------------------------------------------//
// ̈ʒuԂ
// index - ̂̃CfbNX
// step - Runge-Kutta@̃Xebv
    protected function getPosition(index:int,step:int):Position{
      if(index<0) return null;
      return tempPos[index][step];
    }

//--------------------------------------------------------------------//
// xݒ肷
// a - xxNg
// step - Runge - Kutta @̃Xebv
    protected function setAcc(a:Vector.<Number>,step:int):void{
      var bodyCount:int=getBodyCount();
      var ii:int=0;
      for(var i:int=0;i<bodyCount;i++){
      	tempPos[i][step].setAcc(a,ii);
      	ii+=6;
      }
    }

//--------------------------------------------------------------------//
// f[^XV
// dt - ԍݕ
    protected function update(dt:Number):void{
      var bodyCount:int=getBodyCount();
      var dx:Vector3D=new Vector3D();
      var dv:Vector3D=new Vector3D();
      var dp:Vector.<Number>=new Vector.<Number>(4,true);
      var dw:Vector3D=new Vector3D();
      var cdt:Vector.<Number>=Vector.<Number>
      	([dt*C1[0],dt*C1[1],dt*C1[2],dt*C1[3]]);
      for(var i:int=0;i<bodyCount;i++){
      	dx.scaleBy(0.0);
      	dv.scaleBy(0.0);
      	for(var j:int=0;j<4;j++) dp[j]=0.0
      	dw.scaleBy(0.0);
      	for(j=0;j<4;j++){
      	  addVector(dx,tempPos[i][j].v,cdt[j]);
      	  addVector(dv,tempPos[i][j].a,cdt[j]);
      	  var ddp:Vector.<Number>=tempPos[i][j].eulerVel();
      	  for(var k:int=0;k<4;k++) dp[k]+=0.5*cdt[j]*ddp[k];
      	  addVector(dw,tempPos[i][j].dw,cdt[j]);
      	}
      	getBody(i).position().update1(dx,dv,dp,dw);
      }
    }

//--------------------------------------------------------------------//
// xNgɒl
// v1 - xNg
// v2 - xNg
// coef - W
    protected static function addVector(v1:Vector3D,v2:Vector3D,
      	      	      	      	      	coef:Number):void{
      v1.x+=coef*v2.x;
      v1.y+=coef*v2.y;
      v1.z+=coef*v2.z;
    }

//--------------------------------------------------------------------//
// xŃ\[g
    internal function sortLabel(d1:LabeledData,d2:LabeledData):int{
      var label1:int=d1.label();
      var label2:int=d2.label();
      if(label1<label2){
      	return -1;
      }
      else if(label1>label2){
      	return 1;
      }
      else{
      	return 0;
      }
    }

//--------------------------------------------------------------------//
// S\[g
    internal function sortConstraint(c1:Constraint,c2:Constraint):int{
      if(c1.p1()<c2.p1()){
      	return -1;
      }
      else if(c1.p1()>c2.p1()){
      	return 1;
      }
      else if(c1.p2()<c2.p2()){
      	return -1;
      }
      else if(c1.p2()>c2.p2()){
      	return 1;
      }
      else{
      	return 0;
      }
    }
  }
}
